#Standard
import hashlib
import os
import random

from contextlib import contextmanager
from datetime import datetime as dt
from functools import wraps
from urllib.error import HTTPError
from urllib.parse import parse_qs, urlsplit, urlunsplit

#Third-party
import requests

from Crypto import Random
from Crypto.Cipher import AES
from lxml import etree, objectify
from owslib.csw import CatalogueServiceWeb
from sqlalchemy import create_engine, inspect
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.schema import CreateSchema
from sqlalchemy.sql import exists, false as sql_false

#astun
from ishare.shared.core import (join_path_or_uri, get_files_modified,
                                set_qs_parameters)
from models import (Base, CSWService, Instance, MapSource, Layer, Workspace,
                    MetadataRecord, CONFIG_SCHEMA)

BASE_CONFIG_FILE = 'iShareMaps.xml'
APPCONFIG_ENDPOINT = 'intgetdata.aspx?RequestType=AppConfig'
REQUIRED_SETTINGS = ('DB_HOST', 'DB_NAME', 'DB_USER', 'KEY')
OPTIONAL_SETTINGS = ('DB_HOST_PORT', 'DB_PASS', 'WEBSERVICE_URI', 'NAME')
DISPOSE_SESSIONS = False
__AES_KEY__ = 'seriouslydonotusethis'



def path_translator(path):  # pragma: no cover
    return path


def with_instance_config(f):
    @wraps(f)
    def wrapper(session, instance_id, *args, **kwargs):
        mapspur = kwargs.get('mapspur')
        config_path = kwargs.get('config_path')
        if mapspur is None or len(mapspur) == 0 or not config_path:
            instance = session.query(Instance).get(instance_id)
            kwargs['appconfig'], kwargs['mapspur'], kwargs['config_path'] = \
                load_from_webservices(instance.location)
        return f(session, instance_id, *args, **kwargs)
    return wrapper


def get_etree_root(source):
    xml = None
    root = None
    if isinstance(source, etree._Element):
        root = source
    elif isinstance(source, etree._ElementTree):
        xml = source
    else:
        source = path_translator(source)
        xml = etree.parse(source)

    if xml:
        root = xml.getroot()

    return root
    
    
def encrypt_password(plaintext):
    IV = Random.new().read(AES.block_size)
    key = hashlib.md5(__AES_KEY__.encode('utf-8')).hexdigest()
    the_encryptor = AES.new(key.encode('utf-8'), AES.MODE_CFB, IV=IV)
    encrypted = the_encryptor.encrypt(plaintext.encode('utf-8'))
    return encrypted, IV


def get_ishare_info_from_metadata(metadata):
    """
    Expects URI object of the form:
{
    description: "Layer description",
    name: "layername",
    protocol: "OGC:WMS-1.1.1-http-get-map",
    url: "http://host/path/wmservice?MapSource=Astun/Default&"
}"""
    service_uri, mapsource, layer_name = ('',) * 3
    wms = {}
    if metadata.uris and isinstance(metadata.uris, list):
        for uri_obj in metadata.uris:
            protocol = uri_obj.get('protocol')
            if protocol == 'OGC:WMS-1.1.1-http-get-map':
                wms = uri_obj

    wms_url = wms.get('url')
    if wms_url:
        split_uri = urlsplit(wms_url)
        if split_uri.netloc:
            service_uri = urlunsplit((split_uri.scheme, split_uri.netloc,
                                      split_uri.path, '', ''))
            params = dict((k.lower(), v) for k, v in list(parse_qs(
                split_uri.query).items()))
            if params:
                if 'mapsource' in params:
                    mapsource = params['mapsource'][0]
            layer_name = wms.get('name', '')

    return (service_uri, mapsource, layer_name)


def extract_metadata_for_layer(layer, metadata_records, mapsource=None):
    matching_record = None
    other_records = []
    layer_name = layer.name.lower().upper()
    if not mapsource:
        mapsource = layer.mapsources[0]

    mapsource_name = mapsource.name.lower().upper()

    def is_match(record, l_test, ms_test):
        matches = False
        if record.cache:
            uri, ms, l = get_ishare_info_from_metadata(record.cache)
            if all((ms, l)):
                ms_record = ms.lower().upper()
                l_record = l.lower().upper()
                ms_match = (ms_record == ms_test)
                l_match = (l_record == l_test)
                matches = ms_match and l_match
        return matches
    g = ((is_match(r, layer_name, mapsource_name), r) for r in metadata_records)
    for match, record in g:
        if match and not matching_record:
            matching_record = record
        else:
            other_records.append(record)
    return matching_record, other_records


def set_layer_metadata(layer, records, auto_linked):
    unlinked = []
    if layer.metadata:
        m = layer.metadata
        for record in m:
            record.auto_linked = False
            record.link_created = None
            unlinked.append(record)
        del m[:]
    if not isinstance(records, list):
        records = [records]
    for record in records:
        link_layer_metadata(layer, record, auto_linked)

    updated_records = unlinked + records
    return layer, updated_records


def link_layer_metadata(layer, record, auto_linked):
    layer.metadata.append(record)
    record.auto_linked = auto_linked
    record.link_created = dt.today()
    return layer, record


def unlink_mapsource_metadata(session, mapsource, auto_only):
    layers = mapsource.layers
    instance = mapsource.instance
    cswservice = instance.csw
    _check_metadata_cache(session, cswservice)

    updated_layers = []
    updated_metadata = []
    for layer in {l for l in layers if l.metadata}:
        record = layer.metadata[0]
        if auto_only and not record.auto_linked:
            continue
        updated_layer, updated_records = set_layer_metadata(layer, [],
                                                            auto_only)
        updated_layers.append(updated_layer)
        updated_metadata.append(updated_records)

    return tuple(updated_layers), tuple(updated_metadata)


def autolink_mapsource_metadata(session, mapsource):
    instance = mapsource.instance
    cswservice = instance.csw
    _check_metadata_cache(session, cswservice)
    all_metadata = cswservice.metadata

    def generate_matches(layer_list, metadata_list, ms):
        remainder = metadata_list[:]
        for layer in layer_list:
            if not layer.metadata:
                record, remainder = extract_metadata_for_layer(layer,
                                                               remainder,
                                                               mapsource=ms)
                if record:
                    yield set_layer_metadata(layer, record, True)
        return

    updated = tuple(generate_matches(mapsource.layers, all_metadata,
                                     mapsource))
    if updated:
        updated_layers, updated_metadata = list(zip(*updated))
    else:
        updated_layers, updated_metadata = (), ()
    return updated_layers, updated_metadata


def get_csw_credentials(csw):
    username, password = ('',) * 2
    if csw.auth_user:
        username = csw.auth_user
        if csw.auth_pass:
            if csw.auth_pass_iv:
                key = hashlib.md5(__AES_KEY__.encode('utf-8')).hexdigest()
                the_decryptor = AES.new(key.encode('utf-8'), AES.MODE_CFB, csw.auth_pass_iv)
                password = the_decryptor.decrypt(csw.auth_pass)
    return username, password


def combine_xml(base_source, *args):
    def combine_node(base, override):
        known_nodes = {node.tag: node for node in base}
        for node in override:
            if node.tag in known_nodes:
                if not len(node) and node.text:
                    known_nodes[node.tag].text = node.text
                else:
                    combine_node(known_nodes[node.tag], node)
            else:
                known_nodes[node.tag] = node
                base.append(node)

    root = get_etree_root(base_source)

    for override_source in args:
        if override_source == '' or override_source is None:
            continue
        try:
            override_root = get_etree_root(override_source)
        except IOError:
            continue

        combine_node(root, override_root)

    return root


def make_connection_string(provider, user, password, host, port, database):

    if password:
        auth = '{}:{}@'.format(user, password)
    elif user:
        auth = user + '@'
    else:
        auth = ''

    if port:
        address = '{}:{}'.format(host, port)
    else:
        address = host

    connection = '{}://{}{}/{}'.format(provider, auth, address, database)

    return connection


def initialize_db(host, database, user='', password='', port=0,
                  provider='postgresql', instance_uri='', 
                  instance_name='Default'):
    """
    Returns Session class.

    WARNING: Creates schema and tables in database if do not exist!
    """
    if not instance_uri:
        raise ValueError('The "instance_uri" argument must be supplied')
    if not instance_name:
        name = 'Default'
    else:
        name = instance_name

    conn_str = make_connection_string(provider, user, password, host, port,
                                      database)
    engine = create_engine(conn_str, connect_args={'application_name': __file__})
    inspector = inspect(engine)
    schemas = inspector.get_schema_names()
    #if CONFIG_SCHEMA not in schemas:
        #engine.execute(CreateSchema(CONFIG_SCHEMA))
    Base.metadata.create_all(bind=engine)

    Session = scoped_session(sessionmaker())
    Session.configure(bind=engine)

    if instance_uri:
        session = Session()
        try:
            is_instance = session.query(exists().where(
                                        Instance.location ==
                                        instance_uri)).scalar()

            if not is_instance:
                instance = Instance(location=instance_uri, name=name)
                session.add(instance)
                session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()
            session.flush()

    engine.dispose()
    Session.remove()

    return Session


@contextmanager
def scoped_db_session(S):
    """
    Suggested for web requests and similar

    #read parameters, headers, etc.
    with scoped_db_session() as session:
        #do session stuff
    #send response
    """
    session = S()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()
        if DISPOSE_SESSIONS:
            engine = session.get_bind()
            engine.dispose()


def select_cswservice_by_uri(session, uri):
    try:
        csw = session.query(CSWService).filter(
            CSWService.service_uri == uri).one()
    except NoResultFound:
        csw = None
    return csw


def select_cswservice_info(session, csw_id):
    csw = session.query(CSWService).get(csw_id)
    return csw


def select_cswservices(session):
    cswservices = session.query(CSWService).all()
    return cswservices


def select_layer_id(session, layer_name, mapsource_id):
    try:
        layers = session.query(Layer).filter(Layer.mapsources.any(
            MapSource.mapsource_id == mapsource_id))
        layer = layers.filter(Layer.name == layer_name).one()
        layer_id = layer.layer_id
    except NoResultFound:
        layer_id = 0

    return layer_id


def select_layer_info(session, layer_id):
    layer = session.query(Layer).get(layer_id)
    return layer


def select_layers(session, instance_id=0, workspace_id=0, mapsource_id=0,
                  query=''):

    layers = session.query(Layer)
    found = True

    if instance_id:
        layers = layers.filter(Layer.instance_id == instance_id)

    if workspace_id:
        layers = layers.filter(Layer.mapsources.any(
            MapSource.workspaces.any(
                Workspace.workspace_id == workspace_id)))

    if mapsource_id:
        layers = layers.filter(Layer.mapsources.any(
                               MapSource.mapsource_id == mapsource_id))

    if layers.count() == 0 and (instance_id or workspace_id or mapsource_id):
        found = False

    if query:
        layers = layers.filter(Layer.display_name.ilike('%{}%'.format(query)))

    return set(layers.all()), found


def select_mapsource_id(session, mapsource_name, instance_id):
    try:
        mapsources = session.query(MapSource).filter(
            MapSource.instance_id == instance_id)
        mapsource = mapsources.filter(MapSource.name == mapsource_name).one()
        mapsource_id = mapsource.mapsource_id
    except NoResultFound:
        mapsource_id = 0
    return mapsource_id


def select_mapsource_info(session, mapsource_id):
    mapsource = session.query(MapSource).get(mapsource_id)
    return mapsource


def select_mapsources(session, instance_id=0, workspace_id=0):
    mapsources = session.query(MapSource)
    if instance_id:
        mapsources = mapsources.filter(MapSource.instance_id == instance_id)
    if workspace_id:
        mapsources = mapsources.filter(MapSource.workspaces.any(
            Workspace.workspace_id == workspace_id))
    return set(mapsources.all())


def select_workspace_id(session, workspace_name, instance_id):
    try:
        workspaces = session.query(Workspace).filter(
            Workspace.instance_id == instance_id)
        workspace = workspaces.filter(Workspace.name == workspace_name).one()
        workspace_id = workspace.workspace_id
    except NoResultFound:
        workspace_id = 0
    return workspace_id


def select_workspace_info(session, workspace_id):
    workspace = session.query(Workspace).get(workspace_id)
    return workspace


def select_workspaces(session, instance_id=0):
    workspaces = session.query(Workspace)
    if instance_id:
        workspaces = workspaces.filter(
            Workspace.instance_id == instance_id)
    return set(workspaces.all())


def select_instance_id(session, lookup, use_location=False):

    try:
        if use_location:
            instance = session.query(Instance).filter(
                Instance.location == lookup).one()
        else:
            instance = session.query(Instance).filter(
                Instance.name == lookup).one()
        instance_id = instance.instance_id
    except NoResultFound:
        instance_id = 0
    return instance_id


def select_instance_info(session, instance_id):
    instance = session.query(Instance).get(instance_id)
    return instance


def select_metadata_links(session):
    return []


def _check_metadata_cache(session, csw):
    if not csw.cached:
        load_metadata(session, csw.csw_id)
    else:
        cache_delta = dt.today() - csw.cached
        expired = cache_delta.total_seconds() > csw.cache_interval
        if expired:
            load_metadata(session, csw.csw_id)


def select_metadata_records(session, csw_id=0, mapsource_id=0, identifiers=None,
                            auto_only=False):
    metadata_exists = False
    mapsource = None
    records = None
    results = set()

    if mapsource_id:
        mapsource = select_mapsource_info(session, mapsource_id)
    if mapsource and not csw_id:
        csw_id = mapsource.instance.csw_id
    csw = select_cswservice_info(session, csw_id)
    
    if csw:
        _check_metadata_cache(session, csw)
        records = session.query(MetadataRecord).filter(MetadataRecord.csw_id ==
                                                       csw_id)
                                                       
    if records:
        metadata_exists = True
        if mapsource:
            _layers = session.query(Layer.layer_id).filter(Layer.mapsources.contains(mapsource), Layer.metadata).distinct()
            layer_ids = tuple(id for (id,) in _layers.all())
           
            if layer_ids:
                records = records.filter(MetadataRecord.layer_id.in_(layer_ids))
                
            else:
                records = records.filter(sql_false())
                metadata_exists = False
                
        if identifiers:
            records = records.filter(MetadataRecord.identifier.in_(identifiers))
        elif identifiers is not None:
            records = records.filter(sql_false())
            
        if auto_only:
            records = records.filter(MetadataRecord.auto_linked)
            
        results = set(records.all())
        
    return results, metadata_exists


def select_metadata_record(session, record_id):
    record = session.query(MetadataRecord).get(record_id)
    if not record:
        return None

    _check_metadata_cache(session, record.csw)
    checked_record = session.query(MetadataRecord).get(record_id)

    return checked_record


def read_metadata(csw, anytext='', elementSetName='full'):
    csw_kwargs = {}
    records_kwargs = {'esn': elementSetName}
    username, password = get_csw_credentials(csw)
    if username:
        csw_kwargs['username'] = username
        if password:
            csw_kwargs['password'] = password
    csw_service = CatalogueServiceWeb(csw.service_uri, **csw_kwargs)

    if anytext:
        records_kwargs['cql'] = "csw:AnyText LIKE '{}'".format(anytext)

    csw_service.getrecords2(maxrecords=1, **records_kwargs)
    count = csw_service.results.get('matches')
    if count:
        csw_service.getrecords2(maxrecords=count, **records_kwargs)
        results = csw_service.records
    else:
        results = {}
    return results


def read_appconfig_from_ishare(uri, workspace=''):
    if workspace:
        appconfig_uri = set_qs_parameters(uri, {'wor': workspace})
    else:
        appconfig_uri = uri

    appconfig_doc = objectify.parse(appconfig_uri)
    return appconfig_doc.getroot()


def read_appconfig_from_settings(settings_path, mapspur='', workspace_name=''):

    settings_obj = read_settings(settings_path)
    paths_path = settings_obj.paths.attrib.get('file')

    selected = settings_obj.selectedDataConfig
    data_xpath = 'data/config[@name="{}"]'.format(selected)
    data_conf = settings_obj.find(data_xpath)
    data_path = data_conf.get('path')

    return read_appconfig_from_filesystem(paths_path, data_path, mapspur,
                                          workspace_name)


def read_appconfig_from_filesystem(paths_path, data_path, mapspur='',
                                   workspace_name=''):

    paths_tree = get_etree_root(paths_path)

    data_path = path_translator(data_path)
    appconfig = combine_xml(paths_tree, data_path)

    if workspace_name:
        mapspur_folder = ''
        if mapspur and isinstance(mapspur, str):
            mapspur_folder = mapspur
        elif not mapspur:
            mapspur_folder = os.path.join(paths_tree.get('RootPath'), 'config')

        if mapspur_folder:
            mapspur = read_mapspur(mapspur_folder)

        workspace_config = read_workspace_config(mapspur, workspace_name)
        if workspace_config is not None:
            appconfig = combine_xml(appconfig, workspace_config)

    appconfig = objectify.fromstring(etree.tostring(appconfig))

    return appconfig


def read_settings(settings_path):
    settings_path = path_translator(settings_path)
    settings_doc = objectify.parse(settings_path)
    return settings_doc.getroot()


def read_paths(paths_path):
    paths_path = path_translator(paths_path)
    paths_doc = objectify.parse(paths_path)
    return paths_doc.getroot()


def read_paths_from_settings(settings_obj):
    paths_path = settings_obj.paths.attrib.get('file')
    return read_paths(paths_path)


def read_config_path(paths_obj):
    config_path = paths_obj.attrib.get('RootPath')
    return os.path.join(config_path, 'config')


def read_mapspur(folder_path):
    mapspur_path = join_path_or_uri(folder_path, BASE_CONFIG_FILE)
    mapspur_path = path_translator(mapspur_path)
    mapspur_doc = objectify.parse(mapspur_path)
    return mapspur_doc.getroot()


def read_workspace_names(mapspur):
    names = []
    if (mapspur.find('Workspaces') is not None and
            mapspur.Workspaces.find('Workspace') is not None):
        for workspace in mapspur.Workspaces.Workspace:
            names.append(workspace.attrib.get('name'))
    return set(names)


def read_workspace_config(mapspur, workspace_name):
    conf_xpath = 'Workspaces/Workspace[@name="{}"]/ConfigurationDefinition'
    conf_xpath = conf_xpath.format(workspace_name)
    return mapspur.find(conf_xpath)


def read_workspace_map_names(mapspur):
    names = dict()

    for name in read_workspace_names(mapspur):
        map_names = read_map_names(mapspur, workspace=name)
        if map_names:
            names[name] = map_names

    return names


def read_map_names(mapspur, workspace=''):
    names = {}
    mapnames = []
    ws_config = None
    if workspace:
        path = 'Workspace[@name="{}"]'.format(workspace)
        ws_config = mapspur.Workspaces.find(path)
        if ws_config is None:
            raise ValueError('Workspace "{}" not found'.format(workspace))

    if ws_config is not None and len(
            ws_config.findall('MyMapsMapSources/MapName')):
        for mapname in ws_config.MyMapsMapSources.MapName:
            mapnames.append(str(mapname))
    else:
        for mapsource in mapspur.MapSources.MapSource:
            mapnames.append(str(mapsource.MapName))

    for name in mapnames:
        xpath = 'MapSource[MapName/.="{}"]'.format(name)
        result = mapspur.MapSources.xpath(xpath)
        mapsource = result[0]
        display_name = str(mapsource.DisplayName)
        names[name] = display_name

    return names


def read_mapsource(parent_path, mapsource_name):
    parent_path = path_translator(parent_path)
    mapsource_file = join_path_or_uri(parent_path, mapsource_name + '.xml')
    mapsource_doc = objectify.parse(mapsource_file)
    return mapsource_doc.getroot()


def read_mapsources_modified(parent_path, mapsource_names):
    tuples = []
    if parent_path.startswith('http'):
        for name in mapsource_names:
            uri = join_path_or_uri(parent_path, name + '.xml')
            r = requests.head(uri)
            last_modified_header = r.headers.get('Last-Modified')
            last_modified = dt.strptime(last_modified_header,
                                        '%a, %d %b %Y %H:%M:%S %Z')
            tuples.append((name, last_modified))
    else:
        parent_path = path_translator(parent_path)
        tuples = get_files_modified(mapsource_names, relative_to=parent_path,
                                    suffix='.xml', omit_us=True)
    results = {n: d for n, d in tuples}
    return results


def remove_csw(session, instance):
    csw_removed = False
    csw_deleted = False
    csw_uri = ''
    if instance.csw:
        csw = instance.csw
        csw_uri = csw.service_uri
        csw.instances.remove(instance)
        csw_removed = True
        if not csw.instances:
            session.delete(csw)
            csw_deleted = True

    return session, csw_removed, csw_deleted, csw_uri


def disable_csw(csw):
    csw.enabled = False
    return csw


def enable_csw(csw):
    csw.enabled = True
    return csw


def create_csw(session, uri, name, version, metadata_uri_format='', username='',
               password='', cache_interval=0, enabled=False):

    kwargs = {'service_uri': uri, 'name': name, 'version': version,
              'enabled': enabled, 'auth_user': username,
              'cache_interval': cache_interval,
              'metadata_uri_format': metadata_uri_format}
    if username and password:
        kwargs['auth_pass'], kwargs['auth_pass_iv'] = encrypt_password(password)
    csw = CSWService(**kwargs)
    session.add(csw)
    return session, csw

    
def update_csw(csw, username='', password='', cache_interval=0, metadata_uri_format=''):
    csw_user, csw_pass = get_csw_credentials(csw)
    if csw_user != username:
        csw.auth_user = username
    if csw_pass != password:
        csw.auth_pass, csw.auth_pass_iv = encrypt_password(password)
    if cache_interval != csw.cache_interval:
        csw.cache_interval = cache_interval
    if metadata_uri_format != csw.metadata_uri_format:
        csw.metadata_uri_format = metadata_uri_format
    return csw


def assign_csw(instance, csw):
    instance.csw = csw
    return instance


def update_instance_csw(session, instance, appconfig):
    csw = None
    if appconfig.find('CatalogueService/URI'):
        csw_uri = appconfig.CatalogueService.URI.text
        csw_user = ''
        csw_pass = ''
        
        if hasattr(appconfig.CatalogueService, 'CSWUsername'):
            csw_user = appconfig.CatalogueService.CSWUsername.text or ''
            if hasattr(appconfig.CatalogueService, 'CSWPassword'):
                csw_pass = appconfig.CatalogueService.CSWPassword.text or ''          
        elif hasattr(appconfig.CatalogueService, 'Username'):
            csw_user = appconfig.CatalogueService.Username.text or ''
            if hasattr(appconfig.CatalogueService, 'Password'):
                csw_pass = appconfig.CatalogueService.Password.text or ''
        
        if hasattr(appconfig.CatalogueService, 'CacheInterval'):
            cache_interval = appconfig.CatalogueService.CacheInterval.text
        else:
            cache_interval = 0
            
        if hasattr(appconfig.CatalogueService, 'LinkFormat'):
            uri_format = appconfig.CatalogueService.LinkFormat.text
        else:
            uri_format = ''

        try:
            csw_kwargs = {}
            if csw_user:
                csw_kwargs['username'] = csw_user
                if csw_pass:
                    csw_kwargs['password'] = csw_pass
            csw_service = CatalogueServiceWeb(csw_uri, **csw_kwargs)
        except (HTTPError, requests.HTTPError):
            csw = select_cswservice_by_uri(session, csw_uri)
            if csw:
                csw = disable_csw(csw)
                instance = assign_csw(instance, csw)
        else:
            csw = select_cswservice_by_uri(session, csw_uri)
            if csw:
                csw = update_csw(csw, csw_user, csw_pass, cache_interval, uri_format)
                csw = enable_csw(csw)
                if not instance.csw or not instance.csw == csw:
                    instance.csw = csw
            else:
                csw_name = csw_service.identification.title.lower().replace(' ',
                                                                            '_')
                csw_version = csw_service.version
                session, csw = create_csw(session, csw_uri, csw_name,
                                          csw_version, username=csw_user,
                                          password=csw_pass,
                                          cache_interval=cache_interval,
                                          metadata_uri_format=uri_format,
                                          enabled=True)
                instance.csw = csw
            csw.version = csw_service.version
    else:
        session, removed, deleted, uri = remove_csw(session, instance)

    return session, instance, csw


@with_instance_config
def load_mapsource_data(session, instance_id, appconfig=None, mapspur=None,
                        config_path=''):
    names = read_map_names(mapspur)
    map_names = set(names.keys())
    last_updates = read_mapsources_modified(config_path, map_names)

    mapsources = session.query(MapSource).filter(
        MapSource.instance_id == instance_id)
    layers = session.query(Layer).filter(Layer.instance_id == instance_id)
    workspaces = session.query(Workspace).filter(
        Workspace.instance_id == instance_id)

    if map_names:
        ms_removed = mapsources.filter(~MapSource.name.in_(map_names))
        layers_removed = layers.filter(
            Layer.mapsources.any(~MapSource.name.in_(map_names)))
        layers_removed.delete(synchronize_session=False)
        ms_removed.delete(synchronize_session=False)

        ms_new_names = set(map_names)
        ms_existing = mapsources.filter(MapSource.name.in_(map_names))
        if ms_existing:
            for mapsource in ms_existing:
                last_update = last_updates.get(mapsource.name)
                mapsource.last_update = last_update
        ms_existing_names = set(v[0] for v in ms_existing.values('name'))
        if ms_existing_names:
            ms_new_names = ms_new_names - ms_existing_names

        if ms_new_names:
            new = []
            for name in ms_new_names:
                last_update = last_updates.get(name)
                display_name = names.get(name)
                new.append(MapSource(name=name, display_name=display_name,
                                     last_update=last_update,
                                     instance_id=instance_id))
            session.add_all(new)

    ws_names = []
    ws_map_names = read_workspace_map_names(mapspur)
    if ws_map_names:
        ws_names = list(ws_map_names.keys())
        for ws_name in ws_names:
            try:
                ws = workspaces.filter(
                    Workspace.name == ws_name).one()
            except NoResultFound:
                ws = Workspace(instance_id=instance_id, name=ws_name)
                session.add(ws)
            map_names = ws_map_names.get(ws_name)
            if map_names:
                ws_ms = list(mapsources.filter(MapSource.name.in_(map_names)))
            ws.mapsources = ws_ms

    if ws_names:
        ws_removed = workspaces.filter(~Workspace.name.in_(ws_names))
        ws_removed.delete(synchronize_session=False)

    return session


@with_instance_config
def load_layer_data(session, instance_id, appconfig=None, mapspur=None,
                    config_path=''):
    mapsources = session.query(MapSource).filter(
        MapSource.instance_id == instance_id)

    layers = []
    for ms in mapsources.all():
        ms_obj = read_mapsource(config_path, ms.name)
        existing_layers = session.query(Layer).filter(
            Layer.mapsources.contains(ms))
        layer_objects = ms_obj.findall('Layers/LayerGroup/Layer')
        layer_names = []
        for layer_obj in layer_objects:
            layer_name = str(layer_obj.LayerName.text)
            layer_names.append(layer_name)
            try:
                existing_layers.filter(Layer.name == layer_name).filter(
                    Layer.mapsources.contains(ms)).one()
            except NoResultFound:
                if layer_obj.Type in ('base', 'ows'):
                    layer = Layer(name=layer_name, instance_id=instance_id,
                                  mapsources=[ms],
                                  display_name=layer_obj.Name.text)
                    layers.append(layer)
        missing_ids = {l.layer_id for l in existing_layers
                       if l.name not in layer_names}
        if missing_ids:
            missing_layers = session.query(Layer).filter(
                Layer.layer_id.in_(missing_ids))
            if missing_layers.count() > 0:
                missing_layers.delete(synchronize_session=False)

    session.add_all(layers)

    return session


def load_data(session, instance_id):
    session = load_instance_data(session, instance_id)
    session = load_mapsource_data(session, instance_id)
    session = load_layer_data(session, instance_id)
    return session


@with_instance_config
def load_instance_data(session, instance_id, appconfig=None, mapspur=None,
                       config_path=''):
    instance = session.query(Instance).filter(
        Instance.instance_id == instance_id).one()
    session, instance, csw = update_instance_csw(session, instance, appconfig)
    return session


def load_metadata(session, csw_id,  appconfig=None, mapspur=None,
                  config_path=''):
    csw = select_cswservice_info(session, csw_id)
    all_records = read_metadata(csw)
    existing_records = (session.query(MetadataRecord).filter(
        MetadataRecord.csw_id == csw_id))
    existing_idents = [r.identifier for r in existing_records.all()]
    missing_idents = list(set(existing_idents) - set(all_records.keys()))
    missing_records = existing_records.filter(MetadataRecord.identifier.in_(
        missing_idents))
    new_records = []

    def cache(identifier, record):
        if identifier in existing_idents:
            r = existing_records.filter(MetadataRecord.identifier == identifier)[0]
            r.cache = record
        else:
            r = MetadataRecord(identifier=identifier, csw_id=csw_id,
                               cache=record)
            new_records.append(r)
        return r

    try:
        cached_records = list(map(cache, list(all_records.keys()),
                             list(all_records.values())))
        session.add_all(new_records)
    except:
        raise
    else:
        csw.cached = dt.today()

    missing_records.delete(synchronize_session='fetch')
    return session


def load_from_webservices(webservice_base_uri):
    base_uri = str(webservice_base_uri).rstrip('/')
    appconfig_uri = '/'.join((base_uri, APPCONFIG_ENDPOINT))
    appconfig = read_appconfig_from_ishare(appconfig_uri)
    config_path = '/'.join((base_uri, 'config'))
    mapspur = read_mapspur(config_path)
    return appconfig, mapspur, config_path


def load_from_files(settings_path):
    appconfig = read_appconfig_from_settings(settings_path)
    config_path = read_config_path(appconfig)
    mapspur = read_mapspur(config_path)
    return appconfig, mapspur, config_path


def populate_from_config(config):
    if not set(REQUIRED_SETTINGS).issubset(set(config.keys())):
        raise KeyError('Missing settings ({} not all present in {})'.format(
            REQUIRED_SETTINGS,  list(config.keys())))
    global __AES_KEY__
    __AES_KEY__ = config['KEY']
    Session = initialize_db(config['DB_HOST'], config['DB_NAME'],
                            user=config.get('DB_USER'),
                            password=config.get('DB_PASS'),
                            port=config.get('DB_PORT'),
                            instance_uri=config['WEBSERVICE_URI'],
                            instance_name=config.get('NAME', ''))

    with scoped_db_session(Session) as s:
        instance_id = select_instance_id(s, config['WEBSERVICE_URI'],
                                         use_location=True)  
        mapsources = select_mapsources(s, instance_id=instance_id)
        for ms in mapsources:
            _check_metadata_cache(s, ms.instance.csw)
        s = load_data(s, instance_id)
    return Session