from datetime import datetime as dt

from sqlalchemy import (Boolean, Column, DateTime, ForeignKey, Integer,
                        LargeBinary, PickleType, String, Table,
                        UniqueConstraint, inspect)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship
from sqlalchemy.orm.collections import InstrumentedList


#CONFIG_SCHEMA = 'astun_ishareconfig'
CONFIG_SCHEMA = 'ishare_config'

Base = declarative_base()
Base.metadata.schema = CONFIG_SCHEMA


ms_layers_assoc = Table('mapsource_layers', Base.metadata,
                        Column('layer_id', Integer,
                               ForeignKey('layers.layer_id',
                                          ondelete='cascade'),
                               primary_key=True),
                        Column('mapsource_id', Integer,
                               ForeignKey('mapsources.mapsource_id',
                                          ondelete='cascade'),
                               primary_key=True))


ws_ms_assoc = Table('workspace_mapsources', Base.metadata,
                    Column('workspace_id', Integer,
                           ForeignKey('workspaces.workspace_id',
                                      ondelete='cascade'),
                           primary_key=True),
                    Column('mapsource_id', Integer,
                           ForeignKey('mapsources.mapsource_id',
                                      ondelete='cascade'),
                           primary_key=True))


class Mixin(object):
    def _get_attr_names(self):
        inspected = inspect(self)
        return list(inspected.attrs.keys())

    def _get_dict(self, datetime_format='%Y-%m-%d %H:%M:%S'):
        keys = self._get_attr_names()

        d = { '_identity': self._get_id()}

        def minimal_dict(model):
            mini = {'_type': type(model).__name__.lower()}
            if isinstance(model, Mixin):
                id_fields = model._get_id()
                mini['_identity'] = id_fields
                for name in id_fields:
                    mini[name] = getattr(model, name)
            return mini

        for key in keys:
            value = getattr(self, key)
            if isinstance(value, InstrumentedList):
                l = []
                for item in value:
                    if isinstance(item, Base):
                        item = minimal_dict(item)
                    l.append(item)
                d[key] = l
            elif isinstance(value, Base):
                d[key] = minimal_dict(value)
            elif isinstance(value, dt):
                d[key] = value.strftime(datetime_format)
            else:
                d[key] = value

        return d

    @classmethod
    def _get_id(cls):
        inspected = inspect(cls)
        pk = inspected.primary_key
        fields = []
        for col in pk:
            fields.append(col.name)
        return tuple(fields)


class Settings(Base):
    __tablename__ = 'settings'

    key = Column(String, primary_key=True)
    value = Column(String, nullable=False)


class CSWService(Base, Mixin):
    __tablename__ = 'csw_services'

    __sensitive__ = ('auth_pass', 'auth_pass_iv')

    csw_id = Column(Integer, primary_key=True)
    metadata_uri_format = Column(String)
    name = Column(String(100), nullable=False)
    service_uri = Column(String(), nullable=False)
    version = Column(String(8), nullable=False)
    auth_user = Column(String())
    auth_pass = Column(LargeBinary())
    auth_pass_iv = Column(LargeBinary())
    cached = Column(DateTime)
    cache_interval = Column(Integer, nullable=False, default=3600)
    enabled = Column(Boolean, nullable=False, default=False)

    instances = relationship('Instance', backref=backref('csw'))


class Instance(Base, Mixin):
    __tablename__ = 'instances'

    instance_id = Column(Integer, primary_key=True)
    location = Column(String(), unique=True, nullable=False)
    name = Column(String(100), unique=True, nullable=False)
    public_location = Column(String())
    csw_id = Column(Integer, ForeignKey('csw_services.csw_id'))

    layers = relationship('Layer', backref=backref('instance'))
    mapsources = relationship('MapSource', backref=backref('instance'))
    workspaces = relationship('Workspace', backref=backref('instance'))


class Workspace(Base, Mixin):
    __tablename__ = 'workspaces'
    workspace_id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False)
    instance_id = Column(Integer, ForeignKey('instances.instance_id'),
                         nullable=False)
    csw_id = Column(Integer, ForeignKey('csw_services.csw_id'))
    csw = relationship('CSWService', backref=backref('workspaces'))


class MapSource(Base, Mixin):
    __tablename__ = 'mapsources'
    __table_args__ = (UniqueConstraint('name', 'instance_id'),)

    mapsource_id = Column(Integer, primary_key=True)
    name = Column(String(), nullable=False)
    display_name = Column(String, nullable=False)
    last_update = Column(DateTime(), nullable=False)
    instance_id = Column(Integer, ForeignKey('instances.instance_id'))

    workspaces = relationship('Workspace', secondary=ws_ms_assoc,
                              backref=backref('mapsources', cascade='all'),
                              passive_deletes=True)


class Layer(Base, Mixin):
    __tablename__ = 'layers'

    layer_id = Column(Integer, primary_key=True)
    instance_id = Column(Integer, ForeignKey('instances.instance_id'),
                         nullable=False)
    name = Column(String(), nullable=False)
    display_name = Column(String(), nullable=False)

    mapsources = relationship("MapSource", secondary=ms_layers_assoc,
                              backref=backref('layers',
                                              cascade='all, delete-orphan',
                                              single_parent=True),
                              passive_deletes=True)


class MetadataRecord(Base, Mixin):
    __tablename__ = 'metadata_records'

    record_id = Column(Integer, primary_key=True)
    identifier = Column(String(), nullable=False)
    auto_linked = Column(Boolean, nullable=False, default=False)
    link_created = Column(DateTime)
    cache = Column(PickleType)
    csw_id = Column(Integer, ForeignKey('csw_services.csw_id'),
                    nullable=False)
    layer_id = Column(Integer, ForeignKey('layers.layer_id', onupdate="CASCADE",
                                          ondelete="SET NULL"))

    csw = relationship('CSWService', backref=backref('metadata'))
    layer = relationship('Layer', uselist=False,
                         backref=backref('metadata', single_parent=True))
