import configparser
import pickle
import json
import os
import shutil
import sys
import unittest

from collections import OrderedDict
from datetime import datetime
from urllib.error import HTTPError

import lxml

import manager
import webapp

import logging
log = logging.getLogger(__name__)

test_folder = os.path.dirname(os.path.abspath(sys.argv[0]))
test_files_folder = os.path.join(test_folder, 'test_files')
cfg_path = os.path.join(test_folder, 'tests.cfg')
cfg = configparser.ConfigParser()
cfg.read(cfg_path)

manager.DISPOSE_SESSIONS = True

ISHARE_CFG_KEYS = ('appconfig', 'mapspur', 'config_path')
PROVIDER_KNOWN_DB = {'postgresql': 'postgres'}
SCHEMA = 'ishare_config'
MOCK_CSW_URI = 'http://mock.astun.co.uk/services/csw'
MOCK_CSW_TITLE = 'Mock CSW Title'
MOCK_CSW_VERSION = '2.0.2'
MOCK_MAPSOURCE_PARENT_URI = 'http://mock.astun.co.uk/services/ishare/config/'
MOCK_MAPSOURCE_MODIFIED = {'Test/Default': 'Wed, 20 Nov 2015 17:59:46 GMT',
                           'Test/Editing': 'Thu, 29 Oct 2015 09:03:54 GMT',
                           'Test/Education': 'Tue, 18 Nov 2014 12:56:03 GMT',
                           'Test/Test': 'Fri, 20 Nov 2015 17:59:03 GMT'}

RESOURCE_INSTANCE = {'csw': str, 'csw_id': int, 'instance_id': int,
                     'layers': list, 'layers_count': int, 'layers_digest': str,
                     'location': str, 'mapsources': list,
                     'mapsources_count': int, 'mapsources_digest': str,
                     'name': str, 'public_location': str,
                     'resource_identity': list, 'resource_type': str,
                     'resource_uri': str, 'workspaces': list,
                     'workspaces_count': int, 'workspaces_digest': str}

RESOURCE_WORKSPACE = {'csw': str, 'csw_id': int, 'instance': str,
                      'instance_id': int, 'mapsources': list,
                      'mapsources_count': int, 'mapsources_digest': str,
                      'name': str, 'resource_identity': list,
                      'resource_type': str, 'resource_uri': str,
                      'workspace_id': int}

RESOURCE_MAPSOURCE = {'display_name': str, 'instance': str,
                      'instance_id': int, 'last_update': str, 'layers': list,
                      'layers_count': int, 'layers_digest': str,
                      'mapsource_id': int, 'name': str,
                      'resource_identity': list, 'resource_type': str,
                      'resource_uri': str, 'workspaces': list}

RESOURCE_LAYER = {'display_name': str, 'instance': str, 'instance_id': int,
                  'layer_id': int, 'mapsources': list, 'metadata': list,
                  'name': str, 'resource_identity': list, 'resource_type': str,
                  'resource_uri': str}

RESOURCE_CSW = {'auth_user': str, 'cache_interval': int, 'cached': str,
                'csw_id': int, 'enabled': bool, 'instances': list,
                'metadata': list, 'metadata_count': int,
                'metadata_digest':  list, 'name': str,
                'resource_identity': list, 'resource_type': str,
                'resource_uri': str, 'service_uri': str, 'version': str,
                'workspaces': list}

RESOURCE_METADATA = {"resource_identity": list, "layer_id": int, "layer": str,
                     "link_created": str, "record_id": int,
                     "auto_linked": bool, "csw": str, "identifier": str,
                     "csw_id": int, "resource_type": str, "resource_uri": str}
RESOURCE_METADATA_DETAIL = dict(RESOURCE_METADATA)
RESOURCE_METADATA_DETAIL.update({"record": dict})
 
 
def _testify_path(file_name):
    if file_name.startswith('T:\\'):
        file_name = file_name.replace('T:\\', '')

    return os.path.join(test_files_folder, file_name)


class MockRequests(object):
    uris = {}

    def __init__(self):
        for name, lmd in list(MOCK_MAPSOURCE_MODIFIED.items()):
            uri = '{}{}.xml'.format(MOCK_MAPSOURCE_PARENT_URI, name)
            self.uris[uri] = lmd

    def head(self, uri):
        if uri in list(self.uris.keys()):
            response = type('', (), {})()
            response.headers = {'Last-Modified': self.uris.get(uri)}
        else:
            raise HTTPError(uri, 404, 'Mock URI {} is not recognised'.format(
                uri), {}, None)

        return response


class MockCatalogueServiceWeb(object):
    
    with open(_testify_path('metadata/metadata.pkl'),'rb') as f:
        _records = pickle.load(f)
    identification = type('', (), {})()

    def __init__(self, uri):
        self.identification.title = MOCK_CSW_TITLE
        self.version = MOCK_CSW_VERSION
        if uri != MOCK_CSW_URI:
            raise HTTPError(uri, 404, 'Mock URI is not recognised', {}, None)

    def getrecords2(self, maxrecords=10, esn='full', cql=''):
        if cql:
            records = OrderedDict()
            string = cql.strip('csw:AnyText LIKE ').strip("'")
            for identifier, cswrecord in self._records.items():
                if string in cswrecord.xml:
                    records[identifier] = cswrecord
        else:
            records = self._records
        record_count = len(records)
        if maxrecords > record_count:
            maxrecords = record_count

        keys = list(records.keys())[:maxrecords]
        values = list(records.values())[:maxrecords]
        self.records = OrderedDict(list(zip(keys, values)))
        self.results = {'matches': record_count}

class workspaced_mapspur:
    path = ''

    def __enter__(self):
        source_file = _testify_path('config/isharemaps_workspaces.xml')
        mapspur_file = _testify_path('config/isharemaps.xml')
        shutil.copyfile(source_file, mapspur_file)
        self.path = mapspur_file
        return mapspur_file

    def __exit__(self, type_, value, traceback):
        os.remove(self.path)


class simple_mapspur:
    path = ''

    def __enter__(self):
        source_file = _testify_path('config/isharemaps_simple.xml')
        mapspur_file = _testify_path('config/isharemaps.xml')
        shutil.copyfile(source_file, mapspur_file)
        self.path = mapspur_file
        return mapspur_file

    def __exit__(self, type_, value, traceback):
        os.remove(self.path)

    

class IShareConfigBase(unittest.TestCase):

    config = dict(cfg.items('iShareConfig'))


class IShareConfigFiles(IShareConfigBase):

    @classmethod
    def setUpClass(cls):
        manager.CatalogueServiceWeb = MockCatalogueServiceWeb
        manager.requests = MockRequests()
        manager.DISPOSE_SESSIONS = True

    def setUp(self):
        manager.path_translator = _testify_path

    def test_read_settings(self):
        settings = manager.read_settings(self.config['settings_path'])
        self.assertIsInstance(settings, lxml.etree.ElementBase)
        self.assertEqual(settings.paths.get('file'), self.config['paths_path'])

    def test_read_paths_from_settings(self):
        settings = manager.read_settings(self.config['settings_path'])
        paths = manager.read_paths_from_settings(settings)
        self.assertIsInstance(paths, lxml.objectify.ObjectifiedElement)
        self.assertTrue(paths.Web.AutoRefresh)

    def test_read_paths(self):
        paths = manager.read_paths('paths.xml')
        self.assertIsInstance(paths, lxml.objectify.ObjectifiedElement)
        self.assertEqual(manager.read_config_path(paths),
                         self.config['config_path'])

    def test_read_mapspur_with_workspaces(self):
        with workspaced_mapspur():
            mapspur = manager.read_mapspur(self.config['config_path'])
        self.assertIsInstance(mapspur, lxml.objectify.ObjectifiedElement)
        self.assertGreater(len(mapspur.Workspaces.Workspace), 0)
        ws_map_names = manager.read_workspace_map_names(mapspur)
        self.assertEqual(len(ws_map_names), 3)
        self.assertEqual(len(ws_map_names[self.config['workspace_name']]), 2)
        self.assertEqual(list(ws_map_names[self.config['workspace_name']])[
                             1], 'Test/Editing')

    def test_read_mapspur_without_workspaces(self):
        with simple_mapspur():
            mapspur = manager.read_mapspur(self.config['config_path'])
        self.assertIsInstance(mapspur, lxml.objectify.ObjectifiedElement)
        with self.assertRaises(AttributeError):
            count = len(mapspur.Workspaces)

    def test_read_map_names(self):
        with simple_mapspur():
            mapspur = manager.read_mapspur(self.config['config_path'])
            map_names = manager.read_map_names(mapspur)
            self.assertTrue(self.config['mapsource'] in map_names)
            with self.assertRaises(AttributeError):
                map_names = manager.read_map_names(mapspur, 'testing')
        with workspaced_mapspur():
            mapspur = manager.read_mapspur(self.config['config_path'])
            names = manager.read_map_names(mapspur)
            self.assertTrue(self.config['mapsource'] in list(names.keys()))

            names = manager.read_map_names(mapspur, 'testing')
            self.assertFalse(self.config['mapsource'] in list(names.keys()))

            names = manager.read_map_names(mapspur, self.config[
                'workspace_name'])
            self.assertTrue(self.config['mapsource'] in list(names.keys()))

            with self.assertRaises(ValueError):
                map_names = manager.read_map_names(mapspur, 'fake')

    def test_read_mapsource(self):
        ms_name = self.config['mapsource']
        ms = manager.read_mapsource(self.config['config_path'], ms_name)
        self.assertIsInstance(ms, lxml.objectify.ObjectifiedElement)
        self.assertEqual(ms.MapEngine, 'MAPSERVER')

    '''
    def test_read_mapsources_modified_file(self):
        with simple_mapspur():
            mapspur = manager.read_mapspur(self.config['config_path'])
            map_names = manager.read_map_names(mapspur)
        ms_modified = manager.read_mapsources_modified(
            self.config['config_path'], map_names)
        self.assertEqual(len(ms_modified), 4)
        self.assertTrue(self.config['mapsource'] in list(ms_modified.keys()))
        modified_values = [int(v.strip()) for v in
                           self.config['mapsource_modified'].split(',')]
        modified = datetime(*modified_values)
        self.assertEqual(ms_modified.get(self.config['mapsource']), modified)
    '''

    def test_read_mapsources_modified_http(self):
        with simple_mapspur():
            mapspur = manager.read_mapspur(self.config['config_path'])
            map_names = manager.read_map_names(mapspur)
        ms_modified = manager.read_mapsources_modified(
            MOCK_MAPSOURCE_PARENT_URI, map_names)
        self.assertEqual(len(ms_modified), 4)
        self.assertTrue(self.config['mapsource'] in list(ms_modified.keys()))
        modified_values = [int(v.strip()) for v in
                           self.config['mapsource_modified'].split(',')]
        modified = datetime(*modified_values)
        self.assertEqual(ms_modified.get(self.config['mapsource']), modified)

    def test_read_workspace_appconfig(self):
        with workspaced_mapspur():
            appconfig1 = manager.read_appconfig_from_filesystem\
                (self.config['paths_path'], self.config['data_path'],
                 workspace_name=self.config['workspace_name'])
            appconfig2 = manager.read_appconfig_from_filesystem(
                self.config['paths_path'], self.config['data_path'],
                workspace_name=self.config['workspace_name'],
                mapspur=self.config['config_path'])
        self.assertEqual(lxml.etree.tostring(appconfig1),
                         lxml.etree.tostring(appconfig2))
        self.assertIsInstance(appconfig1, lxml.etree.ElementBase)
        self.assertEqual(appconfig1.get('RootPath'), self.config['root_path'])
        self.assertEqual(appconfig1.WebService.SourceRootDir,
                         self.config['root_path'])

    def test_read_appconfig(self):
        appconfig = manager.read_appconfig_from_settings(
            self.config['settings_path'])
        self.assertIsInstance(appconfig, lxml.etree.ElementBase)
        self.assertEqual(appconfig.get('RootPath'), self.config['root_path'])
        self.assertEqual(appconfig.WebService.SourceRootDir,
                         self.config['root_path'])

    def test_combine_xml(self):
        out1 = manager.combine_xml(self.config['paths_path'], '', None,
                                        self.config['data_path'],
                                        self.config['data_path'], 'tests.cfg')
        out2 = manager.combine_xml(self.config['paths_path'],
                                        self.config['data_path'])
        self.assertEqual(lxml.etree.tostring(out1), lxml.etree.tostring(out2))


class IShareConfigDatabase(IShareConfigBase):

    Session = None

    @classmethod
    def _make_engine(cls, db):
        conn_str = manager.make_connection_string(cls.config['provider'],
                                                  cls.config['user'],
                                                  cls.config['password'],
                                                  cls.config['host'],
                                                  cls.config['port'],
                                                  db)
        engine = manager.create_engine(conn_str)
        return engine

    @classmethod
    def setUpClass(cls):
        manager.CatalogueServiceWeb = MockCatalogueServiceWeb
        manager.requests = MockRequests()
        manager.DISPOSE_SESSIONS = True

        manager.path_translator = _testify_path
        engine = cls._make_engine(PROVIDER_KNOWN_DB.get(cls.config['provider']))
        with engine.connect() as conn:
            conn.execute('commit')
            conn.execute('CREATE DATABASE "{}"'.format(cls.config['database']))
        engine.dispose()
        engine = cls._make_engine(cls.config['database'])
        with engine.connect() as conn:
            conn.execute('commit')
            conn.execute('CREATE SCHEMA "{}"'.format(SCHEMA))
        engine.dispose()
        cls.Session = manager.initialize_db(cls.config['host'],
                                            cls.config['database'],
                                            cls.config['user'],
                                            cls.config['password'],
                                            cls.config['port'],
                                            provider=cls.config['provider'],
                                            instance_uri=cls.config[
                                                'instance_uri'])
        return cls.Session

    def setUp(self):
        with simple_mapspur():
            ishare_cfg_values = \
                manager.load_from_files(self.config['settings_path'])
            self.ishare_cfg = dict(list(zip(ISHARE_CFG_KEYS, ishare_cfg_values)))
            self.appconfig = ishare_cfg_values[0]

    @classmethod
    def tearDownClass(cls):
        engine = cls._make_engine(PROVIDER_KNOWN_DB.get(cls.config['provider']))
        with engine.connect() as conn:
            conn.execute('commit')
            conn.execute('DROP DATABASE "{}"'.format(cls.config['database']))


    def test_make_connection_string(self):
        conn_str = 'postgresql://hostname/db'
        self.assertEqual(manager.make_connection_string('postgresql',
                                                             '', '',
                                                             'hostname', 0,
                                                             'db'), conn_str)
        conn_str = 'postgresql://user@hostname/db'
        self.assertEqual(manager.make_connection_string('postgresql',
                                                             'user', '',
                                                             'hostname', 0,
                                                             'db'), conn_str)
        conn_str = 'postgresql://user:pass@hostname/db'
        self.assertEqual(manager.make_connection_string('postgresql',
                                                             'user', 'pass',
                                                             'hostname', 0,
                                                             'db'), conn_str)
        conn_str = 'postgresql://user@hostname:1234/db'
        self.assertEqual(manager.make_connection_string('postgresql',
                                                             'user', '',
                                                             'hostname', 1234,
                                                             'db'), conn_str)

    def test_initialize_db(self):
        with self.assertRaises(ValueError):
            manager.initialize_db(self.config['host'],
                                       self.config['database'],
                                       self.config['user'],
                                       self.config['password'],
                                       self.config['port'],
                                       provider=self.config['provider'])

    def test_load_mapsources_simple(self):
        with manager.scoped_db_session(self.Session) as session:
            session = manager.load_mapsource_data(session, 1,
                                                       **self.ishare_cfg)
            mapsources = manager.select_mapsources(session, 1)
            self.assertEqual(len(mapsources), 4)

    def test_load_layer_data_simple(self):
        with manager.scoped_db_session(self.Session) as session:
            session = manager.load_mapsource_data(session, 1,
                                                       **self.ishare_cfg)
            session = manager.load_layer_data(session, 1,
                                                   **self.ishare_cfg)

    def _load_mapsources_workspace(self, session):
        with workspaced_mapspur():
                appconfig, mapspur, config_path =\
                    manager.load_from_files(self.config['settings_path'])
        ishare_cfg_values = appconfig, mapspur, config_path
        ishare_cfg = dict(list(zip(ISHARE_CFG_KEYS, ishare_cfg_values)))
        session = manager.load_mapsource_data(session, 1, **ishare_cfg)
        workspace_id = manager.select_workspace_id(session,
                                                        self.config[
                                                        'workspace_name'],
                                                        1)
        mapsources = manager.select_mapsources(session, 1,
                                                    workspace_id)
        return session, appconfig, mapspur, config_path, mapsources

    def test_load_mapsources_workspace(self):
        with manager.scoped_db_session(self.Session) as session:
            _, _, _, _, mapsources = self._load_mapsources_workspace(session)
            self.assertEqual(len(mapsources), 2)

    def test_load_layers_workspace(self):
        with manager.scoped_db_session(self.Session) as session:
            session, appconfig, mapspur, config_path, mapsources = \
                self._load_mapsources_workspace(session)
            ms = next(iter(mapsources))
            layer1 = manager.Layer(name='test1', instance_id=1,
                                   mapsources=[ms],
                                   display_name='TEST LAYER 1')
            layer2 = manager.Layer(name='test2', instance_id=1,
                                   mapsources=[ms],
                                   display_name='TEST LAYER 2')

            session.add_all([layer1, layer2])
            test_layers, __ = manager.select_layers(session, 1,
                                                         query='TEST')
            self.assertEqual(len(test_layers), 2)
            session = manager.load_layer_data(session, 1,
                                                   **self.ishare_cfg)
            test_layers, __ = manager.select_layers(session, 1,
                                                         query='TEST')

            all_layers = session.query(manager.Layer).all()
            self.assertGreater(len(all_layers), 0)
            self.assertEqual(len(test_layers), 0)

    def test_select_csw_functions(self):
        uri = self.config['instance_uri']
        version = '2.0.2'
        with manager.scoped_db_session(self.Session) as session:
            session, csw = manager.create_csw(session, uri, 'TEST',
                                              version, enabled=True)
            with self.assertRaises(TypeError):
                session, csw = manager.create_csw()
            with self.assertRaises(TypeError):
                session, csw = manager.create_csw(session, uri, 'TEST')
        with manager.scoped_db_session(self.Session) as session:
            csw = manager.select_cswservice_by_uri(session, uri)
            self.assertIsInstance(csw, manager.CSWService)
            self.assertEqual(csw.service_uri, uri)
            self.assertNotEqual(csw.service_uri, '')
            self.assertEqual(csw.name, 'TEST')
            self.assertNotEqual(csw.name, ' TEST ')
            self.assertTrue(csw.enabled)

    def test_load_instance_data(self):
        with manager.scoped_db_session(self.Session) as session:
            self.appconfig.CatalogueService.URI = MOCK_CSW_URI
            instance = manager.select_instance_info(session, 1)
            if instance.csw:
                manager.remove_csw(session, instance)

            session = manager.load_instance_data(session, 1, **self.ishare_cfg)
            instance = manager.select_instance_info(session, 1)
            csw = instance.csw
            try:
                self.assertTrue(csw.enabled)
                self.assertEqual(csw.service_uri, MOCK_CSW_URI)
                self.assertEqual(csw.version, MOCK_CSW_VERSION)
                self.assertNotEqual(csw.name, MOCK_CSW_TITLE)
                self.assertEqual(csw.name, MOCK_CSW_TITLE.lower().replace(' ',
                                                                          '_'))
            except AssertionError:  # pragma: no cover
                manager.remove_csw(session, instance)

    def test_update_instance_csw_existing(self):
        with manager.scoped_db_session(self.Session) as session:
            session, existing_csw = manager.create_csw(session,
                                                       MOCK_CSW_URI,
                                                       MOCK_CSW_TITLE,
                                                       '2.0.2')
            self.appconfig.CatalogueService.URI = MOCK_CSW_URI
            instance = manager.select_instance_info(session, 1)
            if instance.csw:
                manager.remove_csw(session, instance)
            session, instance, csw = \
                manager.update_instance_csw(session, instance,
                                                 self.appconfig)
            self.assertEqual(existing_csw, csw)

    '''
    def test_update_instance_csw_existing_invalid(self):
        invalid_uri = 'NOT A URI'
        with manager.scoped_db_session(self.Session) as session:
            session, existing_csw = manager.create_csw(session, invalid_uri,
                                                       'existing', '2.0.2')
            self.appconfig.CatalogueService.URI = invalid_uri
            instance = manager.select_instance_info(session, 1)
            if instance.csw:
                manager.remove_csw(session, instance)
            session, instance, csw =\
                manager.update_instance_csw(session, instance,
                                                 self.appconfig)
            self.assertEqual(existing_csw, csw)

    def test_update_instance_csw_new_invalid(self):
        invalid_uri = 'NOT A URI'
        with manager.scoped_db_session(self.Session) as session:
            self.appconfig.CatalogueService.URI = invalid_uri
            instance = manager.select_instance_info(session, 1)
            if instance.csw:
                manager.remove_csw(session, instance)
            session, instance, csw =\
                manager.update_instance_csw(session, instance,
                                                 self.appconfig)
            self.assertIsNone(csw)
            self.assertIsNone(instance.csw)
    '''

    def test_update_instance_csw_no_uri(self):
        with manager.scoped_db_session(self.Session) as session:
            session, existing_csw = manager.create_csw(session,
                                                            MOCK_CSW_URI,
                                                            MOCK_CSW_TITLE,
                                                            MOCK_CSW_VERSION)
            instance = manager.select_instance_info(session, 1)
            manager.assign_csw(instance, existing_csw)
            del self.appconfig.CatalogueService.URI
            session, instance, csw =\
                manager.update_instance_csw(session, instance,
                                                 self.appconfig)
            self.assertIsNone(csw)
            self.assertIsNone(instance.csw)


class WebAppTestCase(unittest.TestCase):

    Session = None

    def _is_JSON(self, string):
        try:
            js = json.loads(string)
            self.assertTrue(True)
        except ValueError:  # pragma: no cover
            self.fail('Not a valid JSON string')
        else:
            return js

    def _check_resource(self, checkee, description, collection=''):
        def go_compare(r, d):
            self.assertCountEqual(list(r.keys()), list(d.keys()))

        if collection:
            checklist = checkee.get(collection)
        else:
            checklist = [checkee]

        for resource in checklist:
            go_compare(resource, description)

    def _get_and_validate(self, path, description,
                          collection='', expected_status=200):
        resp = self.app.get(path)
        if resp.status_code != expected_status:
            msg = 'Unexpected status: {} ({})'.format(resp.status, path)
            raise ValueError({'msg': msg, 'status_code': resp.status_code})
        resource = self._is_JSON(resp.data)
        self._check_resource(resource, description, collection)
        return resource

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        manager.path_translator = _testify_path
        with workspaced_mapspur():
            cls.appconfig, cls.mapspur, cls.config_path = \
                manager.load_from_files(IShareConfigDatabase.config[
                    'settings_path'])
        cls.appconfig.CatalogueService.URI = MOCK_CSW_URI
        cfg = {'appconfig': cls.appconfig, 'mapspur': cls.mapspur,
               'config_path': cls.config_path}
        webapp.app.DBSession = IShareConfigDatabase.setUpClass()
        webapp.app.config['TESTING'] = True
        cls.app = webapp.app.test_client()
        with manager.scoped_db_session(webapp.app.DBSession) as session:
            session = manager.load_instance_data(session, 1, **cfg)
            session = manager.load_mapsource_data(session, 1, **cfg)
            session = manager.load_layer_data(session, 1, **cfg)
            session = manager.load_metadata(session, 1, **cfg)


    @classmethod
    def tearDownClass(cls):
        IShareConfigDatabase.tearDownClass()

    def test_instance(self):
        instance0 = self._get_and_validate('/instance/1', RESOURCE_INSTANCE)
        self.assertEqual(instance0.get('name'), 'Default')
        instance1 = self._get_and_validate('/instance/Default',
                                           RESOURCE_INSTANCE)
        self.assertEqual(instance1.get('instance_id'), 1)
        self.assertEqual(instance0, instance1)

    def test_workspace_collection(self):
        workspace_collection = self._get_and_validate('/instance/1/workspace',
                                                      RESOURCE_WORKSPACE,
                                                      collection='workspaces')
        workspace_list = workspace_collection.get('workspaces')
        self.assertGreater(len(workspace_list), 0)
        for ws in workspace_list:
            self.assertEqual(ws.get('instance_id'), 1)
        with self.assertRaises(ValueError):
            self._get_and_validate('/instance/9999/workspace',
                                   RESOURCE_WORKSPACE, collection='workspaces')

    def test_workspace(self):
        workspace = self._get_and_validate('/instance/1/workspace/public',
                                           RESOURCE_WORKSPACE)
        with self.assertRaises(ValueError):
            self._get_and_validate('/instance/1/workspace/',
                                   RESOURCE_WORKSPACE)

    def test_mapsource_collection(self):
        mapsource_collection = self._get_and_validate('/instance/1/mapsource',
                                                      RESOURCE_MAPSOURCE,
                                                      collection='mapsources')
        mapsource_list = mapsource_collection.get('mapsources')
        self.assertGreater(len(mapsource_list), 0)
        for ms in mapsource_list:
            self.assertEqual(ms.get('instance_id'), 1)

        instance = self._get_and_validate('/instance/1', RESOURCE_INSTANCE)
        workspaces = instance.get('workspaces')
        if workspaces:
            ws_path = workspaces[0]
            coll_path = '{}/mapsource'.format(ws_path)
            self._get_and_validate(coll_path, RESOURCE_MAPSOURCE,
                                   collection='mapsources')

        with self.assertRaises(ValueError):
            self._get_and_validate('/instance/9999/mapsource',
                                   RESOURCE_MAPSOURCE, collection='mapsources')

    def test_mapsource(self):
        mapsource0 = self._get_and_validate(
            '/instance/1/mapsource/Test/Default', RESOURCE_MAPSOURCE)
        ms_id = mapsource0.get('mapsource_id')
        mapsource1 = self._get_and_validate('/mapsource/{}'.format(ms_id),
                                            RESOURCE_MAPSOURCE)
        self.assertEqual(mapsource0, mapsource1)

        instance = self._get_and_validate('/instance/1', RESOURCE_INSTANCE)
        workspaces = instance.get('workspaces')
        if workspaces:
            ws_path = workspaces[0]
            ms_path = '{}/mapsource/Test/Default'.format(ws_path)
            self._get_and_validate(ms_path, RESOURCE_MAPSOURCE)

        with self.assertRaises(ValueError):
            self._get_and_validate('/instance/1/mapsource/Not/Real',
                                   RESOURCE_MAPSOURCE)

    def test_layer_collection(self):
        layer_collection = self._get_and_validate('/instance/1/layer',
                                                  RESOURCE_LAYER,
                                                  collection='layers')
        layer_list = layer_collection.get('layers')
        self.assertGreater(len(layer_list), 0)
        with self.assertRaises(ValueError):
            self._get_and_validate('/instance/9999/layer', RESOURCE_LAYER,
                                   collection='layers')

    def test_layer(self):
        mapsource = self._get_and_validate(
            '/instance/1/mapsource/Test/Default', RESOURCE_MAPSOURCE)
        ms_id = mapsource.get('mapsource_id')
        layer0_path = '/mapsource/{}/layer/transport_bus'.format(ms_id)
        layer0 = self._get_and_validate(layer0_path, RESOURCE_LAYER)
        layer_id = layer0.get('layer_id')
        layer1_path = '/layer/{}'.format(layer_id)
        layer1 = self._get_and_validate(layer1_path, RESOURCE_LAYER)
        self.assertEqual(layer0, layer1)
        with self.assertRaises(ValueError):
            self._get_and_validate('/layer/9999', RESOURCE_LAYER)
    
    '''
    def test_cswservice_collection(self):
        csw_collection = self._get_and_validate('/cswservice', RESOURCE_CSW,
                                                collection='cswservices')
        csw_list = csw_collection.get('cswservices')
        self.assertGreater(len(csw_list), 0)
        uri = self.appconfig.CatalogueService.URI
        found_cfg_csw_uri = any(csw['service_uri'] == uri for csw in
                                csw_list)
        self.assertTrue(found_cfg_csw_uri)

    def test_cswservice(self):
        cswservice = self._get_and_validate('/instance/1', RESOURCE_INSTANCE)
        uri_from_conf = self.appconfig.CatalogueService.URI
        csw_path = '/cswservice/{}'.format(cswservice.get('csw_id'))
        csw0 = self._get_and_validate(csw_path, RESOURCE_CSW)
        csw1 = self._get_and_validate(cswservice.get('csw'), RESOURCE_CSW)
        self.assertEqual(csw0, csw1)
        self.assertEqual(csw0.get('service_uri'), uri_from_conf)
    '''

    def test_metadata(self):
        record = self._get_and_validate('metadata/1', RESOURCE_METADATA_DETAIL)
        self.assertFalse(record.get('auto_linked'))
        self.assertEqual(record.get('identifier'),
                         '83756971-2116-42d8-a48a-9c6800516a45')
        self.assertFalse(record.get('auto_linked'))
        record = self._get_and_validate('metadata/904',
                                        RESOURCE_METADATA_DETAIL)
        cache = record.get('record')
        name = cache.get('uris')[0].get('name')
        self.assertEqual(name, 'county_councillors')
        with self.assertRaises(ValueError):
            self._get_and_validate('metadata/99999', RESOURCE_METADATA)

    def test_metadata_csw_collection(self):
        metadata_collection = self._get_and_validate('cswservice/1/metadata',
                                                     RESOURCE_METADATA,
                                                     collection='metadata')
        metadata_list = metadata_collection.get('metadata')
        self.assertEqual(len(metadata_list), 965)
        metadata_collection = self._get_and_validate(
            'cswservice/1/metadata?record_detail=extract',
            RESOURCE_METADATA_DETAIL, collection='metadata')
        metadata_collection = self._get_and_validate(
            'cswservice/1/metadata?record_detail=full',
            RESOURCE_METADATA_DETAIL, collection='metadata')

    def test_metadata_autolink(self):
        autolink_uri = 'mapsource/1/metadata/auto'
        resp = self.app.get(autolink_uri)
        self.assertEqual(resp.status_code, 404)
        resp = self.app.post(autolink_uri)
        self.assertEqual(resp.status_code, 200)
        resp = self.app.post(autolink_uri)
        self.assertEqual(resp.status_code, 422)
        coll = self._get_and_validate(autolink_uri, RESOURCE_METADATA,
                                      collection='metadata')
        auto_metadata = coll.get('metadata')
        self.assertEqual(len(auto_metadata), 20)
        record = self._get_and_validate('metadata/904',
                                        RESOURCE_METADATA_DETAIL)
        self.assertEqual(record.get('layer_id'), 6)
        nonexistent_mapsource_uri = 'mapsource/0/metadata/auto'
        resp = self.app.get(nonexistent_mapsource_uri)
        self.assertEqual(resp.status_code, 404)
        resp = self.app.post(nonexistent_mapsource_uri)
        self.assertEqual(resp.status_code, 422)
        resp = self.app.delete(autolink_uri)
        self.assertEqual(resp.status_code, 204)
        resp = self.app.get(autolink_uri)
        self.assertEqual(resp.status_code, 404)
        resp = self.app.delete(autolink_uri)
        self.assertEqual(resp.status_code, 404)
        incorrect_metadata_path = 'mapsource/1/metadata/_test'
        resp = self.app.get(incorrect_metadata_path)
        self.assertEqual(resp.status_code, 404)
        resp = self.app.post(incorrect_metadata_path)
        self.assertEqual(resp.status_code, 405)
        invalid_metadata_path = 'mapsource/1/metadata/not_valid_keyword'
        resp = self.app.get(invalid_metadata_path)
        self.assertEqual(resp.status_code, 404)

    def test_metadata_layer(self):
        record_id = 904
        layer_uri = 'layer/6'
        layer_metadata_uri = layer_uri + '/metadata'
        metadata_ref = json.dumps({'metadata_id': record_id})
        layer = self._get_and_validate(layer_uri, RESOURCE_LAYER)
        self.assertFalse(layer.get('metadata'))
        resp = self.app.get(layer_metadata_uri)
        self.assertEqual(resp.status_code, 404)
        resp = self.app.put(layer_metadata_uri, data=metadata_ref,
                            content_type='application/json')
        self.assertEqual(resp.status_code, 200)
        coll = self._get_and_validate(layer_metadata_uri, RESOURCE_METADATA,
                                      collection='metadata')
        layer_metadata = coll.get('metadata')
        metadata = layer_metadata[0]
        self.assertEqual(metadata.get('record_id'), record_id)
        resp = self.app.put(layer_metadata_uri, data=metadata_ref,
                            content_type='application/json')
        self.assertEqual(resp.status_code, 409)
        resp = self.app.delete(layer_metadata_uri)
        self.assertEqual(resp.status_code, 204)
        resp = self.app.delete(layer_metadata_uri)
        self.assertEqual(resp.status_code, 404)
        resp = self.app.get(layer_metadata_uri)
        self.assertEqual(resp.status_code, 404)
        faulty_json = json.dumps({'incorrect': 'this is wrong'})
        resp = self.app.put(layer_metadata_uri, data=faulty_json,
                            content_type='application/json')
        self.assertEqual(resp.status_code, 400)
        not_a_record = json.dumps({'metadata_id': 9999})
        resp = self.app.put(layer_metadata_uri, data=not_a_record,
                            content_type='application/json')
        self.assertEqual(resp.status_code, 404)

    def test_metadata_searches(self):
        csw_search_uri = 'cswservice/1/metadata?metadata_search=councillor'
        coll = self._get_and_validate(csw_search_uri, RESOURCE_METADATA,
                                      collection='metadata')
        results = coll.get('metadata')
        self.assertEqual(len(results), 4)
        nonmatching_csw_search = 'cswservice/1/metadata?metadata_search=nomatch'
        coll = self._get_and_validate(nonmatching_csw_search, RESOURCE_METADATA,
                                      collection='metadata')
        results = coll.get('metadata')
        self.assertEqual(len(results), 0)
        autolink_uri = 'mapsource/1/metadata/auto'
        resp = self.app.post(autolink_uri)
        self.assertEqual(resp.status_code, 200)
        mapsource_search_uri = 'mapsource/1/metadata?metadata_search=school'
        coll = self._get_and_validate(mapsource_search_uri, RESOURCE_METADATA,
                                      collection='metadata')
        results = coll.get('metadata')
        self.assertEqual(len(results), 4)
        nonmatching_csw_search = 'mapsource/1/metadata?metadata_search=nomatch'
        coll = self._get_and_validate(nonmatching_csw_search, RESOURCE_METADATA,
                                      collection='metadata')
        results = coll.get('metadata')
        self.assertEqual(len(results), 0)
        self.app.delete(autolink_uri)

    def test_metadata_layer_search(self):
        autolink_uri = 'mapsource/1/metadata/auto'
        resp = self.app.post(autolink_uri)
        self.assertEqual(resp.status_code, 200)
        search_layer_uri = 'mapsource/1/metadata?layer_search=st'
        coll = self._get_and_validate(search_layer_uri, RESOURCE_METADATA,
                                      collection='metadata')
        layer_results = coll.get('metadata')
        self.assertEqual(len(layer_results), 4)
        search_layer_uri = 'mapsource/1/metadata?metadata_search=port'
        coll = self._get_and_validate(search_layer_uri, RESOURCE_METADATA,
                                      collection='metadata')
        metadata_results = coll.get('metadata')
        self.assertEqual(len(metadata_results), 6)
        search_combo_uri = 'mapsource/1/metadata?metadata_search=port&layer_search=st'
        coll = self._get_and_validate(search_combo_uri, RESOURCE_METADATA,
                                      collection='metadata')
        combo_results = coll.get('metadata')
        self.assertEqual(len(combo_results), 2)
        layer_idents = set(r['identifier'] for r in layer_results)
        metadata_idents = set(r['identifier'] for r in metadata_results)
        ident_intersection = layer_idents.intersection(metadata_idents)
        combo_idents = set(r['identifier'] for r in combo_results)
        self.assertEqual(ident_intersection, combo_idents)
        self.app.delete(autolink_uri)

    def test_load_config_from_env(self):
        webapp.__OPTIONAL_SETTINGS__ = webapp.__OPTIONAL_SETTINGS__ + (
            'CHECK_KEY',)
        all_settings = (webapp.__REQUIRED_SETTINGS__ +
                        webapp.__OPTIONAL_SETTINGS__)
        check = {}
        for key in all_settings:
            if key == 'CHECK_KEY':
                os.environ[key] = check[key] = 'CHECK_VALUE'
            elif webapp.app.config.get(key):
                os.environ[key] = check[key] = str(webapp.app.config[key])
        updated_app = webapp.load_config(webapp.app)
        for key in all_settings:
            self.assertEqual(updated_app.config.get(key), check.get(key))




if __name__ == '__main__':
    unittest.main()