import argparse
import hashlib
import importlib
import inspect
import logging
import os
import sys
from configparser import ConfigParser


class DynamicDataSource(object):
    _name = None
    _display_name = None
    _identity = ()
    _id = None


    @classmethod
    def to_dataset_xml(cls, name, column_names, data):
        """
        Generates 'DataSet' XML given records and column_names

        Arguments:
            data: iterator returning dict-like objects,
                each representing a single record
            column_names: sequence of names corresponding to
                keys in the record object
            name: name to be given to each record node

        """
        records = []
        for record in data:
            columns = []
            for col in column_names:
                try:
                    val = record[col]
                except AttributeError:
                    val = ''
                column_xml = '<{col}><![CDATA[{val}]]></{col}>'.format(col=col, val=val)
                columns.append(column_xml)
            content = ''.join(columns)
            record_xml = '<{name}>{content}</{name}>'.format(name=name,
                                                             content=content)
            records.append(record_xml)
        dataset_xml = '<DataSet>{}</DataSet>'.format(''.join(records))
        return dataset_xml

    @classmethod
    def generate_id(cls, identity):
        return hashlib.md5(identity.encode('utf-8')).hexdigest()


    def _set_private_attributes(self):
        attributes = ('name', 'display_name', 'identity')
        class_name = type(self).__name__
        has_overrides = self._config.has_section('Overrides')
        for attribute in attributes:
            _attr = '_' + attribute
            value = None
            if has_overrides:
                value = self._config.get('Overrides', attribute)
            if not value:
                __ATTR = '_{}__{}'.format(class_name, attribute.upper())
                value = getattr(self, __ATTR, getattr(self, _attr, None))
            setattr(self, _attr, value)

    def _get_path(self):
        class_file = inspect.getfile(self.__class__)
        path = os.path.dirname(os.path.realpath(class_file))
        return path


    def __init__(self):
        defaults = {'name': '', 'display_name': '', 'identity': '',
                    'unique_field': ''}
        self._config = ConfigParser(defaults)
        path = self._get_path()
        self._config.read(os.path.join(path, 'plugin.cfg'))
        self._set_private_attributes()

    def query(self, field, value):
        raise NotImplementedError

    def preview(self):
        raise NotImplementedError

    def test(self):
        field = self._config.get('Test', 'field', fallback='')
        value = self._config.get('Test', 'value', fallback='')
        return self.query(field, value)

    @property
    def fields(self):
        raise NotImplementedError

    @property
    def id(self):
        if not self._id:
            identity_values = [self._name, ]
            for setting in self._identity:
                value = self._config.get('Settings', setting, fallback='')
                identity_values.append(value)
            self._id = DynamicDataSource.generate_id('\n'.join(identity_values))
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def display_name(self):
        return self._display_name

    @property
    def unique_field(self):
        name = None
        if self._config.has_section('Settings'):
            name = self._config.get('Settings', 'unique_field', fallback='')
        return name

    @property
    def path(self):
        return self._get_path()


def try_import(module_dir):
    module = None
    try:
        module = import_dynamic_datasource(module_dir)
    except ImportError as err:
        logging.error('DataSource load of {} failed due to import error'.format(
                      module_dir))
        logging.error(err)
    
    return module


def load_dynamic_datasources_under_folder(folder):
    """
    :param folder: The path to the folder to scan
    :return: A list of DynamicDataSources found in immediate subfolders of
        'folder'
    """
    path = os.path.realpath(folder)
    datasources = {}
    folders = [os.path.join(path, name) for name in os.listdir(path)
               if os.path.isdir(os.path.join(path, name))]
    modules = map(try_import, folders)
    modules = filter(None, modules)
    for module in modules:
        try:
            datasource = module.DataSource()
        except Exception as err:
            logging.error('DataSource load failed from {}'.format(
                          module.__name__))
            logging.error(err)
        else:
            datasources[datasource.id] = datasource
    return datasources


def load_dynamic_datasources(locations):
    all_datasources = {}
    for location in locations:
        datasources = load_dynamic_datasources_under_folder(location)
        all_ids = set(all_datasources.keys())
        ids = set(datasources.keys())
        new = ids.difference(all_ids)
        for ds_id in new:
            all_datasources[ds_id] = datasources[ds_id]
    return all_datasources


def flat_datasources(datasources):
    for id, ds in datasources.items():
        try:
            yield {'id': id, 'name': ds.name,
                   'display_name': ds.display_name,
                   'path': ds.path, 'unique_field': ds.unique_field,
                   'fields': ','.join(ds.fields.split('\n'))}
        except NotImplementedError as err:
            logging.error('DataSource {} not implemented correctly'.format(
                          ds.__module__ ))
            logging.error(err)
            #skip datasources that don't work
            continue
        except Exception as err:
            logging.error('DataSource {} failure'.format(ds.__module__))
            logging.error(err)
            continue

def import_dynamic_datasource(module_dir):
    module = None
    parent, folder = os.path.split(os.path.realpath(module_dir))
    sys.path.append(parent)
    try:
        module = importlib.import_module(folder)
    finally:
        sys.path.remove(parent)

    if not hasattr(module, 'DataSource'):
        raise ImportError(f"No DataSource found in module {folder}")
    import ishare.isharedata.lib
    if not issubclass(module.DataSource, ishare.isharedata.lib.DynamicDataSource.DynamicDataSource):
        raise ImportError(f"DataSource in {folder} is not instance of DynamicDataSource")

    return module

def load_dynamic_datasource(location):
    module = import_dynamic_datasource(location)
    datasource = module.DataSource()
    return datasource


def output_datasource_list(args):
    datasources = load_dynamic_datasources(args.location)
    lines = []
    output = DynamicDataSource.to_dataset_xml('DataSource',
                                              ['id', 'path', 'name',
                                               'display_name', 'unique_field',
                                               'fields'],
                                              flat_datasources(datasources))
    if args.outfile:
        outfile = os.path.realpath(args.outfile)
        print(outfile)
        with open(outfile, 'w') as f:
            f.write(output)
    else:
        print(output)


def output_datasource_call(args):
    datasource = load_dynamic_datasource(args.location)
    output = ''
    if args.test:
        output = datasource.test()
    elif args.preview:
        output = datasource.preview()
    elif args.id:
        output = datasource.id
    elif args.unique_field:
        output = datasource.unique_field
    elif args.name:
        output = datasource.name
    elif args.display_name:
        output = datasource.display_name
    elif args.fields:
        output = datasource.fields
    elif args.query:
        field, value = args.query
        output = datasource.query(field, value)

    print(output)


def main():
    logfile = __file__ + '.log'
    logging.basicConfig(filename=logfile, level=logging.WARNING)
    parser = argparse.ArgumentParser(description='Dynamic DataSource framework')
    subs = parser.add_subparsers()
    call_parser = subs.add_parser('call', help='Run commands on a given'
                                  ' datasource')

    list_parser = subs.add_parser('list', help='List datasources in specified'
                                  ' locations')
    list_parser.add_argument('location', nargs='+', action='store')
    list_parser.add_argument('-o', '--output', dest='outfile', action='store')
    list_parser.set_defaults(action=output_datasource_list)

    call_parser.add_argument('location', action='store')
    call_parser.set_defaults(action=output_datasource_call)
    action_group = call_parser.add_mutually_exclusive_group(required=True)
    action_group.add_argument('--query', dest='query', nargs=2,
                              metavar=('FIELD', 'VALUE'))
    actions = ('test', 'preview', 'id', 'unique_field', 'fields', 'name',
               'display_name')
    for action in actions:
        action_group.add_argument('--{}'.format(action), action='store_true')

    args = parser.parse_args()
    args.action(args)


if __name__ == '__main__':
    main()





